%%
clear all
close all

%% Set various parameters of the optimizations
% basePar is the physical parameters of the model (masses, etc)
basePar = getAWalkerParams;

% baseSoln is a single starting state to work from.
% baseSoln = [dphi0 dphi1 k imp];
baseSoln = [-2.1191 -1.1688 2.1059 0.4621].';

% targetVels and targetStepLens are the desired walking speeds and step lengths
% to solve for
targetVels = 0.8:0.1:2.5;
targetStepLens = 0.6:0.05:1.2;

% errorThreshold is the cyclic error threshold that needs to be reached to
% consider a particular solution found.
errorThreshold = 1e-5;

% resultsFile is a mat file to save results to as they're found.
resultsFile = 'AWalkerResults.mat';

%% Initialization
nVels = length(targetVels);
nLens = length(targetStepLens);

% Compute the leg angle for each desired step length.
initLegAngles = zeros([nLens,1]);
for sind = 1:nLens
    initLegAngles(sind) = fzero(@(legang) stepLengthError(legang,targetStepLens(sind),basePar),pi/4);
end

% Initialize solutions
solns = repmat(baseSoln,[1,nVels,nLens]);

% Initialize errors
errors = zeros([length(baseSoln),nVels,nLens]);
absErrors = zeros([nVels,nLens]);
for vind = 1:nVels
    cVel = targetVels(vind);
    for sind = 1:nLens
        cLen = targetStepLens(sind);
        cAngle = initLegAngles(sind);

        cX = solns(:,vind,sind);
        err = cyclicWalkError(cX,basePar,cLen,cVel,cAngle);
        errors(:,vind,sind) = err;
        absErrors(vind,sind) = max(abs(err));
    end
end

%% Begin searching for cyclic gaits
fitWindowSize = [5 5];
fitWindowMid = (fitWindowSize - 1)/2 + 1;
[ fitdv fitds ] = ind2sub(fitWindowSize,1:prod(fitWindowSize));
fitdv = fitdv - fitWindowMid(1);
fitds = fitds - fitWindowMid(2);

searchWindowSize = [3 3];
searchWindowMid = (searchWindowSize - 1)/2 + 1;
[ searchdv searchds ] = ind2sub(searchWindowSize,1:prod(searchWindowSize));
searchdv = searchdv - searchWindowMid(1);
searchds = searchds - searchWindowMid(2);

numSeqFails = 0;
while sum(absErrors(:) < errorThreshold) < numel(absErrors) && numSeqFails < 2*(numel(absErrors)-sum(absErrors(:) < errorThreshold))
    [B,IX] = sort(absErrors(:));
    if B(1) > errorThreshold
        svind = ceil(nVels*nLens*rand);
        [vind,sind] = ind2sub([nVels,nLens],svind);
    else
        for i = randperm(sum(B<errorThreshold))
            [gvind,gsind] = ind2sub([nVels,nLens],IX(i));
            found = false;
            for j = randperm(length(searchdv))
                vind = gvind + searchdv(j);
                sind = gsind + searchds(j);
                if vind < 1 || vind > nVels
                    continue;
                elseif sind < 1 || sind > nLens
                    continue;
                elseif absErrors(vind,sind) > errorThreshold
                    found = true;
                    break;
                end
            end
            if found
                break;
            end
        end
    end


    stepLen = targetStepLens(sind);
    initLegAngle = initLegAngles(sind);
    vel = targetVels(vind);

    if absErrors(vind,sind) < errorThreshold
        continue;
    end

    vind
    sind

    
    gvind = vind + fitdv;
    gsind = sind + fitds;

    goodWinInds = gvind >= 1 & gvind <= nVels & gsind >= 1 & gsind <= nLens;
    gvind = gvind(goodWinInds);
    gsind = gsind(goodWinInds);

    linInds = sub2ind([nVels nLens],gvind,gsind);
    decentPoints = absErrors < 2*errorThreshold;
    linInds = linInds(decentPoints(linInds));
    [gvind,gsind] = ind2sub([nVels nLens],linInds);

    dvel = targetVels(gvind) - targetVels(vind);
    dstep = targetStepLens(gsind) - targetStepLens(sind);
    W = diag(exp(-0.5*(dvel.^2 + dstep.^2))./(errorThreshold + absErrors(linInds)));
    d = reshape(solns(:,linInds).',[length(linInds),size(solns,1)]);
    X = [ones(length(dvel),1) dvel.' dstep.' (dvel.*dstep).' ];

    if size(X,1) > size(X,2)
        % this fit is only interesting if it's over determined, otherwise
        % it will fit exactly and only return our input point.
        sFit = pinv(W*X)*(W*d);
        sFitErrs = X*sFit - d;
        s0 = sFit(1,:).' + chol(0.001*eye(size(solns,1)) + cov(sFitErrs))*randn([size(solns,1),1]);
    elseif size(X,1) > 1
        s0 = solns(:,linInds(ceil(rand*length(linInds)))) + chol(cov(d))*randn([size(solns,1),1]);
    else
        s0 = solns(:,vind,sind) + 0.5*chol(cov(solns(:,absErrors < 10*errorThreshold).'))*randn([size(solns,1),1]);
    end
    
    [s1, error, initLegAngle] = findWalkParameters(basePar,stepLen,vel,s0,initLegAngle);
    E = max(abs(error));
    if E < absErrors(vind,sind)
        solns(:,vind,sind) = s1;
        errors(:,vind,sind) = error;
        absErrors(vind,sind) = E;
        numSeqFails = 0;
    else
        numSeqFails = numSeqFails + 1;
    end

    sum(absErrors(:) < errorThreshold)/numel(absErrors)
    save(resultsFile);
end

%% Load Results
load(resultsFile);

% Compute starting and ending states
initStates = zeros([4,nVels,nLens]);
endStates = zeros([4,nVels,nLens]);
controlParams = zeros([2,nVels,nLens]);
for vind = 1:nVels
    vel = targetVels(vind);
    for sind = 1:nLens
        stepLen = targetStepLens(sind);
        initLegAngle = initLegAngles(sind);

        soln = solns(:,vind,sind);

        [error,xinit,par,xnext,ts,xs] = cyclicWalkError(soln,basePar,stepLen,vel,initLegAngle);
        initStates(:,vind,sind) = xinit;
        endStates(:,vind,sind) = xs(end,:).';
        controlParams(:,vind,sind) = [par.k;par.imp];
    end
end

%% Display Results
cmap = 'hsv';
validPts = absErrors < errorThreshold;
velsMat = repmat(targetVels',[1,nLens]);
stepLensMat = repmat(targetStepLens,[nVels,1]);

validStepLens = squeeze(stepLensMat(validPts));
validVels = squeeze(velsMat(validPts));
validVelStepTri = delaunay(validStepLens, validVels);

validEndPose1s = squeeze(endStates(1,validPts));
validEndPose2s = squeeze(endStates(2,validPts));
validEndVel1s = squeeze(endStates(3,validPts));
validEndVel2s = squeeze(endStates(4,validPts));
validPoses = squeeze(initStates(1,validPts));
validVel1s = squeeze(initStates(3,validPts));
validVel2s = squeeze(initStates(4,validPts));
validSprings = squeeze(controlParams(1,validPts));
validImps = squeeze(controlParams(2,validPts));

figure(1), clf;
colormap(cmap);
trimesh(validVelStepTri, validStepLens, validVels, validEndVel1s);
xlabel('Step Length'), ylabel('Velocity'), title('Phi1 Vel');

figure(2), clf;
colormap(cmap);
trimesh(validVelStepTri, validStepLens, validVels, validEndVel2s);
xlabel('Step Length'), ylabel('Velocity'), title('Phi2 Vel');

figure(3), clf;
colormap(cmap);
trimesh(validVelStepTri, validStepLens, validVels, validSprings);
xlabel('Step Length'), ylabel('Speed');
title('Spring Constant');

figure(4), clf;
colormap(cmap);
trimesh(validVelStepTri, validStepLens, validVels, validImps);
xlabel('Step Length'), ylabel('Speed');
title('Impulse');

figure(6), clf;
subplot(1,2,1), hist(controlParams(2,:));
title('Impulse distribution');
subplot(1,2,2), hist(controlParams(1,:));
title('Spring Constant distribution');

figure(10), clf;
mesh(targetStepLens,targetVels,absErrors);
xlabel('Step Length'), ylabel('Velocity'), title('Max(Abs(err))');

figure(11), clf;
image(256*absErrors/max(absErrors(:)));
xlabel('Step Length Index'), ylabel('Velocity Index'), title('Max(Abs(err))');

%% Animate the walk of a single speed/step length
sind = 5;
vind = 5;
soln = solns(:,vind,sind);
legAngle = initLegAngles(sind);
par = basePar;
par.k = abs(soln(3));
par.imp = abs(soln(4));
x0 = [ legAngle; -legAngle; soln(1:2) ];
[xnext,tcontact,t,x] = simulateAWalker([0,5], x0, par, 'a');
animateWalker(par,x,5)
